(load "oops-syntax.scm")

;; Split LST into elements, for which predicate PRED is true and
;; elements, for which PRED is false.
(define (oops:%split-on pred lst)
  (let loop ((true '()) (false '()) (rest lst))
    (cond ((null? rest)
           (values true false))
          ((pred (car rest))
           (loop (cons (car rest) true) false (cdr rest)))
          (else
           (loop true (cons (car rest) false) (cdr rest))))))

;; Like SPLIT-ON, but preserve the order of elements.
(define (oops:%stable-split-on pred lst)
  (receive (true false) (oops:%split-on pred lst)
    (values (reverse! true) (reverse! false))))

;; Quicksort the elements in LST, using LESS as a comparison function
;; and SPLIT-ON as split function.
(define (oops:%quicksort less split-on lst)
  (if (or (null? lst) '()
          (null? (cdr lst)))
      lst
      (let ((pred (lambda (elt) (less elt (car lst)))))
        (receive (before after) (split-on pred (cdr lst))
          (append (oops:%quicksort less split-on before)
                  (list (car lst))
                  (oops:%quicksort less split-on after))))))

;; Sort the list LST.
(define (oops:%sort less lst)
  (oops:%quicksort less oops:%split-on lst))

;; Sort the list LST, preserving the relative order of elements.
(define (oops:%stable-sort less lst)
  (oops:%quicksort less oops:%stable-split-on lst))

;; Create an object.  An object is a vector with three elements: 0 -
;; parents list, 1 - ancestors list and 2 - associative slot/value
;; list
(define (oops:%make-object parents . args)
  (letrec ((ancestors
            (lambda (x) (cons x (append-map ancestors (vector-ref x 0)))))
           (obj (vector parents #f (list))))
    (vector-set! obj 1
                 (oops:%stable-sort (lambda (x y) (member x (vector-ref y 0)))
                                    (delete-duplicates (ancestors obj))))
    (apply oops:%tell (append (list obj 'initialize) args))
    obj))


;; Set the slot NAME in object SELF to the value VAL
(define (oops:%slot-set! self name val)
  (let* ((slots (vector-ref self 2))
         (slot (assq name slots)))
    (if slot
        (set-cdr! slot val)
        (vector-set! self 2 (cons (cons name val) slots)))))

;; Get the value of the slot NAME in object SELF.  Return a pair of
;; values, the second indicating boolean sucess/failure
(define (oops:%slot-get2 self name)
  (let* ((slots (vector-ref self 2))
         (slot (assq name slots)))
    (if slot
        (values (cdr slot) #t)
        (values #f #f))))

;; Get the value of the slot NAME in object SELF.  Return #f if
;; there's no such slot.
(define (oops:%slot-get self name)
  (let* ((slots (vector-ref self 2))
         (slot (assq name slots)))
    (if slot (cdr slot) #f)))

;; Call FN on all the elements of LST and return the result of the
;; first successful call.  The function FN returns two values, the
;; second one indicating boolean success/failure
(define (oops:%find2 fn lst)
  (let loop ((rest lst))
    (if (null? rest)
        (values #f #f #f)
        (receive (val win) (fn (car rest))
          (if win
              (values #t val (cdr rest))
              (loop (cdr rest)))))))

;; Find a slot named NAME in SELF and its ancestors, or if SELF is a
;; list, in the members of SELF.
(define (oops:%rget self name)
  (let ((ancestors (if (vector? self) (vector-ref self 1) (cdr self))))
    (oops:%find2 (lambda (x) (oops:%slot-get2 x name)) ancestors)))

;; Get the value of the slot NAME from SELF and its ancestors, or if
;; SELF is a list, in the members of SELF.
(define (oops:%slot-ref self name)
  (receive (win val next) (oops:%rget self name) val))

(define (oops:%tell self message . args)
  (receive (win proc next) (oops:%rget self message)
    (if (not win)
        #f
        (if (vector? self)
            (apply proc (append (list self (cons self next)) args))
            (apply proc (append (list (car self) (cons (car self) next)) args))))))
